// TEView.h : interface of the CTEditView class
//


#pragma once

#include "text.h"

class CTEditDoc;

class CTEditView: public CView
{
protected: // create from serialization only
    CTEditView();
    DECLARE_DYNCREATE(CTEditView)

// Attributes
public:
    CTEditDoc* GetDocument() const;
    UINT lastCharTyped;

    int cursorX, cursorY;
    bool inputDisabled;

    int windowX, windowY;
    int windowWidth, windowHeight;

    int leftMargin, topMargin;
    int rightMargin, bottomMargin;

    bool fontParametersSet;
    int charWidth, charHeight;     // Width, height of a font character
    int leading;        // Interline skip
    int dx, dy;

    TextLine endOfText;
    CFont textFont;
    bool fontLoaded;

    CBrush bgBrush;
    CBrush cursorBgBrush;

    // Status line information
    const char* message;

    // Keyboard status
    bool m_bShiftOn;
    bool m_bControlOn;

// Operations
public:
    void LoadFont();
    bool ProcessKeydownMessage(UINT nChar, UINT nFlags);
    void DrawCursor(
        int cx, int cy,
        bool cursorOn = true
    );
    void DrawCursor(
        int cx, int cy,
        bool cursorOn,
        CDC* pDC
    );
    void PreProcessCommand();
    void PostProcessCommand();
    void RedrawTextRectangle(int x, int y, int w, int h);

    // Scrolling methods
    void ScrollToCursor();
    void ScrollLeft(int n);
    void ScrollRight(int n);
    void ScrollUp(int n);
    void ScrollDown(int n);

    // Command description
    struct CommandDsc
    {
        UINT virtualKey;    // Virtual key (see "Winuser.h")
        UINT state;         // State of Shift and Control
        bool change;        // A command changes a text
        void (CTEditView::*method)(); // Pointer to method processing a command
    };

    // Table of commands
    static const struct CommandDsc editorCommands[];

    // Methods that implement editor commands
    // Cursor movement
    void onDown();      // Down arrow
    void onUp();        // Up arrow
    void onLeft();      // Left arrow
    void onRight();     // Right arrow
    void onHome();      // To the beginning of line
    void onEnd();       // To the end of line
    void onTextBeg();   // To the beginning of text
    void onTextEnd();   // To the end of text
    void onTab();       // Tabulation
    void onTabLeft();   // Tabulation to the left
    void onInsertTab(); // Insert a tabulation (several spaces)
    void onPageUp();    // Page Up
    void onPageDown();  // Page Down

    // "Horizontal" (inline) text changing commands
    void onBackSpace(); // Delete previous character
    void onDelete();    // Delete current character
    void onInsert();    // Insert a space
    void onCharTyped(); // Insert a character typed

    // "Vertical" text changing commands
    void onDeleteLine();   // Delete a complete line
    void onInsertLine();   // Insert an empty above the current
    void onEnter();        // Divide a current line in two pieces

    // Save file
    void onSave();      // Save a text in a file
    void onSaveAs();    // ...not implemented yet...

    void onQuit();      // Ask user, if text was changed, save, and quit
    void close();       // Close editor (quit without questions)

    bool processQuit(); // Process the "Quit" request. Returns "true"
                        // if window should be closed

    // Create new text / open new file
    void onNew();
    void onOpen();

// Overrides
public:
    virtual void OnDraw(CDC* pDC);  // overridden to draw this view
    virtual BOOL PreCreateWindow(CREATESTRUCT& cs);
protected:
    virtual BOOL OnPreparePrinting(CPrintInfo* pInfo);
    virtual void OnBeginPrinting(CDC* pDC, CPrintInfo* pInfo);
    virtual void OnEndPrinting(CDC* pDC, CPrintInfo* pInfo);

// Implementation
public:
    virtual ~CTEditView();
#ifdef _DEBUG
    virtual void AssertValid() const;
    virtual void Dump(CDumpContext& dc) const;
#endif

protected:

// Generated message map functions
protected:
    DECLARE_MESSAGE_MAP()
public:
    virtual void OnInitialUpdate();
    afx_msg void OnChar(UINT nChar, UINT nRepCnt, UINT nFlags);
    afx_msg void OnFileOpen();
    afx_msg void OnFileNew();
    afx_msg void OnFileSave();
    afx_msg void OnUpdateFileSave(CCmdUI *pCmdUI);
    afx_msg void OnFileSaveAs();
    afx_msg void OnUpdateFileSaveAs(CCmdUI *pCmdUI);
    afx_msg void OnSize(UINT nType, int cx, int cy);
    afx_msg BOOL OnEraseBkgnd(CDC* pDC);
protected:
    virtual void OnActivateFrame(UINT nState, CFrameWnd* pDeactivateFrame);
public:
    afx_msg void OnLButtonDown(UINT nFlags, CPoint point);
};

#ifndef _DEBUG  // debug version in TEView.cpp
inline CTEditDoc* CTEditView::GetDocument() const
   { return reinterpret_cast<CTEditDoc*>(m_pDocument); }
#endif
